from typing import Callable

from _jsonrpc import Wrapper
from celestia.types import Namespace
from celestia.types.header import ExtendedHeader
from celestia.types.rawshare import (
    ExtendedDataSquare,
    NamespaceData,
    SampleCoords,
    GetRangeResult,
    RawSample,
)


class ShareClient(Wrapper):
    """Client for interacting with Celestia's Share API."""

    async def get_eds(
        self, height: int, *, deserializer: Callable | None = None
    ) -> ExtendedDataSquare:
        """Gets the full EDS identified by the given extended header.

        Args:
            height (int): The block height.
            deserializer (Callable | None): Custom deserializer. Defaults to :meth:`~celestia.types.share.ExtendedDataSquare.deserializer`.

        Returns:
            ExtendedDataSquare: The retrieved EDS object.
        """

        deserializer = (
            deserializer if deserializer is not None else ExtendedDataSquare.deserializer
        )

        return await self._rpc.call("share.GetEDS", (height,), deserializer)

    async def get_namespace_data(
        self, height: int, namespace: Namespace, *, deserializer: Callable | None = None
    ) -> list[NamespaceData]:
        """Gets all shares from an EDS within the given namespace. Shares are returned
        in a row-by-row order if the namespace spans multiple rows.

        Args:
            height (int): The block height.
            namespace (Namespace): The namespace identifier.
            deserializer (Callable | None): Custom deserializer. Defaults to None.

        Returns:
            list[NamespaceData]: A list of NamespaceData objects or [] if not found.
        """

        def deserializer_(result):
            if result is not None:
                return [NamespaceData(**data) for data in result]
            else:
                return []

        deserializer = deserializer if deserializer is not None else deserializer_

        return await self._rpc.call(
            "share.GetNamespaceData", (height, Namespace(namespace)), deserializer
        )

    async def get_range(
        self, height: int, start: int, end: int, *, deserializer: Callable | None = None
    ) -> GetRangeResult:
        """Gets a list of shares and their corresponding proof.

        Args:
            height (int): The block height.
            start (int): The starting index.
            end (int): The ending index.
            deserializer (Callable | None): Custom deserializer. Defaults to :meth:`~celestia.types.share.GetRangeResult.deserializer`.

        Returns:
            GetRangeResult: The retrieved range result containing shares and proof.
        """

        deserializer = deserializer if deserializer is not None else GetRangeResult.deserializer

        return await self._rpc.call("share.GetRange", (height, start, end), deserializer)

    async def get_samples(
        self,
        header: ExtendedHeader,
        indices: list[SampleCoords],
        *,
        deserializer: Callable | None = None,
    ) -> list[str] | list[RawSample]:
        """Gets sample for given indices.

        Args:
            header (ExtendedHeader): The extended header.
            indices (list[SampleCoords]): A list of sample coordinates.

        Returns:
            list[str]: A list of retrieved samples or [] if not found.
        """

        def default_deserializer(value: list[str | dict] | None):
            if value is None:
                return []
            return [
                (
                    RawSample(**item)  # is the structure in version > 0.20.4
                    if isinstance(item, dict)
                    else item
                )
                for item in value
            ]

        return await self._rpc.call(
            "share.GetSamples",
            (
                header,
                indices,
            ),
            deserializer=(deserializer or default_deserializer),
        )

    async def get_share(self, height: int, row: int, col: int) -> str:
        """Gets a Share by coordinates in EDS.

        Args:
            height (int): The block height.
            row (int): The row index.
            col (int): The column index.

        Returns:
            str: The retrieved share.
        """
        return await self._rpc.call(
            "share.GetShare",
            (
                height,
                row,
                col,
            ),
        )

    async def get_available(self, height: int) -> bool:
        """Subjectively validates if Shares committed to the given ExtendedHeader are available on the Network.

        Args:
            height (int): The block height.

        Returns:
            bool: True if shares are available, False otherwise.
        """
        return await self._rpc.call("share.SharesAvailable", (height,))



import logging
import os.path
import subprocess
from time import sleep

from tests.docker import Container

TESTNET_PATH = os.path.dirname(os.path.abspath(__file__))


def start_testnet() -> bool:
    args = ["docker", "compose", "-f", f"{TESTNET_PATH}/testnet/docker-compose.yml", "up", "-d"]
    proc = subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    if proc.returncode == 0:
        return True
    logging.error(proc.stderr.decode("utf8"))
    return False


def stop_testnet() -> bool:
    args = ["docker", "compose", "-f", f"{TESTNET_PATH}/testnet/docker-compose.yml", "down"]
    proc = subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    if proc.returncode == 0:
        return True
    logging.error(proc.stderr.decode("utf8"))
    return False


def get_auth_token(container: Container) -> str | None:
    args = [
        "docker",
        "exec",
        "-i",
        container.id,
        "celestia",
        container.image,
        "auth",
        "admin",
        "--p2p.network",
        "private",
    ]
    cnt = 10
    while cnt:
        proc = subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if proc.returncode == 0:
            token = proc.stdout.decode("utf8").strip()
            return token
        sleep(1)
        cnt -= 1
        if not cnt:
            logging.error(proc.stderr.decode("utf8"))



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "pyCelestia"
copyright = "2025, Alesh, Tnmoxa"
author = "Alesh, Tnmoxa"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
]

napoleon_google_docstring = True
autodoc_typehints = "none"
viewcode_follow_imported_members = True

templates_path = ["_templates"]
exclude_patterns = []

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
}

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_static_path = ["_static"]

html_theme = "sphinx_rtd_theme"
html_theme_options = {"navigation_depth": 2}



import asyncio
import json
import sys
import typing as t
import uuid
from asyncio import Future
from collections import deque
from collections.abc import AsyncGenerator
from json import JSONEncoder

from ajsonrpc.core import JSONRPC20Response, JSONRPC20Request

from .abc import RPCExecutor, Transport, logger

if sys.version_info[:2] <= (3, 10):
    from async_timeout import timeout as asyncio_timeout

    asyncio.timeout = asyncio_timeout


class RPC(RPCExecutor):
    """RPC encoder / executor / decoder"""

    def __init__(
        self,
        transport: Transport,
        json_encoder: t.Type[JSONEncoder] | None = None,
        timeout: float = 180,
    ):
        self.timeout = timeout
        self.transport = transport
        self.transport.on_message = self.on_transport_response
        self.transport.on_close = self.on_transport_close
        self._pending = dict()  # type: dict[str, Future]
        self._subscriptions = dict()  # type: dict[str, deque[t.Any]]
        self._json_encoder = json_encoder or JSONEncoder

    def on_transport_response(self, message: str):
        message = json.loads(message)
        if "method" in message:
            subscription_id, item = message["params"]
            subscription = self._subscriptions.get(subscription_id, None)
            if subscription is not None:
                subscription.append(item)
        else:
            response = JSONRPC20Response(result=False)
            response.body = message
            if future := self._pending.get(response.id):
                if response.error is not None:
                    error_body = getattr(response.error, "body", None)
                    error_message = error_body.get("message", None).lower() if error_body else None
                    if found := [
                        key for key in self.transport.errors_map.keys() if key in error_message
                    ]:
                        exc_class = self.transport.errors_map[found[0]]
                        future.set_exception(exc_class(error_message))
                    elif error_message is None or error_body is None:
                        future.set_exception(ConnectionError("RPC failed; undefined error"))
                    else:
                        future.set_exception(
                            ConnectionError(f"RPC failed; {error_message}", response.error)
                        )
                else:
                    future.set_result(response.result)
            else:
                logger.warning("Received message with unexpected ID.")

    def on_transport_close(self, exc: Exception = None):
        if exc:
            e = ConnectionError(f"RPC failed; transport closed by error {exc}")
            exc = e.with_traceback(exc.__traceback__)

        for future in self._pending.values():
            if exc:
                future.set_exception(exc)
            else:
                future.set_exception(ConnectionError("RPC failed; transport closed"))

        for id in tuple(self._subscriptions.keys()):
            subscription = self._subscriptions.pop(id)
            if exc:
                subscription.append(exc)

    async def call(
        self,
        method: str,
        params: tuple[t.Any, ...] = None,
        deserializer: t.Callable[[t.Any], t.Any] = None,
    ) -> t.Any | None:
        params = params or ()
        deserializer = deserializer or (lambda a: a)
        id = str(uuid.uuid4())
        request = JSONRPC20Request(method, params, id)
        await self.transport.send(json.dumps(request.body, cls=self._json_encoder))
        future = self._pending[id] = Future()
        future.add_done_callback(lambda _: self._pending.pop(id, None))
        async with asyncio.timeout(self.timeout):
            result = await future
            return deserializer(result)

    async def iter(
        self,
        method: str,
        params: tuple[t.Any, ...] = None,
        deserializer: t.Callable[[t.Any], t.Any] = None,
    ) -> AsyncGenerator[t.Any]:
        deserializer = deserializer or (lambda a: a)
        subscription_id = await self.call(method, params)
        try:
            subscription = self._subscriptions[subscription_id] = deque()
            while subscription_id in self._subscriptions:
                if len(subscription):
                    item = subscription.popleft()
                    if isinstance(item, Exception):
                        raise item
                    yield deserializer(item)
                else:
                    await asyncio.sleep(0.1)
        finally:
            self._subscriptions.pop(subscription_id, None)



import logging
import typing as t
from abc import ABC, abstractmethod
from collections.abc import AsyncGenerator

logger = logging.Logger("WS JSON-RPC")


class Transport(ABC):
    on_message: t.Callable[[bytes | str], None]
    on_close: t.Callable[[Exception | None], None]

    @abstractmethod
    async def send(self, message: str) -> None:
        """Send a message to the connection."""

    @property
    @abstractmethod
    def errors_map(self) -> dict[str, t.Type[Exception]]:
        """Return a mapping of RPC error message (part of the message) to error classes to be raised."""


class RPCExecutor(ABC):

    @abstractmethod
    async def call(
        self,
        method: str,
        params: tuple[t.Any, ...] = None,
        deserializer: t.Callable[[t.Any], t.Any] = None,
    ) -> t.Any | None:
        """This method must implement calling an RPC method and returning the result."""

    @abstractmethod
    async def iter(
        self,
        method: str,
        params: tuple[t.Any, ...] = None,
        deserializer: t.Callable[[t.Any], t.Any] = None,
    ) -> AsyncGenerator[t.Any]:
        """This method must implement creating an `RPC` subscription and returning
        an asynchronous iterator that returns the incoming subscription results.
        """


class Wrapper:
    def __init__(self, rpc: RPCExecutor):
        self._rpc = rpc



import typing as t
from base64 import b64decode, b64encode

from celestia._celestia import types as ext  # noqa


class Base64(bytes):
    """Represents a byte string that supports Base64 encoding and decoding.

    This class ensures that the stored data is always in bytes and provides
    Base64 encoding/decoding when converting to and from strings.
    """

    def __new__(cls, value: str | bytes):
        if isinstance(value, str):
            value = b64decode(value)
        if value is None:
            return None
        return super().__new__(cls, value)

    def __str__(self) -> str:
        return b64encode(self).decode("ascii")

    @classmethod
    def ensure_type(cls, value):
        """Ensures the value is an instance of Base64.

        Args:
            value (str | bytes | Base64): The value to convert.

        Returns:
            Base64: A valid Base64 instance.
        """
        if isinstance(value, cls):
            return value
        return cls(value)


class Namespace(Base64):
    """Represents a Celestia namespace.

    A namespace is a unique identifier for blobs stored on the Celestia network.
    It is used to segregate data based on different applications or use cases.
    """

    def __new__(cls, value: str | bytes):
        value = super().__new__(cls, value)
        value = ext.normalize_namespace(value)
        return super().__new__(cls, value)


class Commitment(Base64):
    """Represents a Celestia blob commitment.

    A commitment is a cryptographic proof that ensures data integrity and allows
    verification of whether a blob has been correctly included in a block.
    """


class TxConfig(t.TypedDict):
    """Represents a transaction configuration for submitting transactions to Celestia.

    Attributes:
        signer_address (str | None): The address of the transaction signer.
        is_gas_price_set (bool | None): Whether a custom gas price is set.
        key_name (str | None): The name of the key used for signing.
        gas_price (float | None): The gas price for the transaction.
        gas (int | None): The amount of gas to use.
        fee_granter_address (str | None): Address of the fee granter (if applicable).
    """

    signer_address: bool | str | None
    is_gas_price_set: bool | None
    key_name: str | None
    gas_price: float | None
    gas: int | None
    fee_granter_address: str | None


class Address(Base64):
    """Represents a address that supports Base64 encoding and decoding.

    This class ensures that the stored address is always in bytes and provides
    Base64 encoding/decoding when converting to and from strings.
    """

    def __new__(cls, value: str | bytes):
        if isinstance(value, str) and value.startswith("celestia"):
            value = ext.address2bytes(value)
        return super().__new__(cls, value)



import asyncio
import typing as t
from contextlib import asynccontextmanager, AbstractAsyncContextManager
from json import JSONEncoder
from urllib.parse import urlparse

from websockets import Headers
from websockets.asyncio.client import connect, ClientConnection

from .abc import Transport as AbcTransport, Wrapper
from .executor import RPC


class Client:
    """WS JSON-RPC Client"""

    def __init__(self, url: str, **options: t.Any):
        pr = urlparse(url)
        if pr.scheme not in ("ws", "wss"):
            raise ValueError("Unsupported URL scheme, must be ws or wss")
        self.options = dict(options, url=url)

    def connect(
        self,
        errors_map: dict[str, t.Type[Exception]] = None,
        additional_headers: Headers | t.Mapping[str, str] | t.Iterable[tuple[str, str]] = None,
        json_encoder: t.Type[JSONEncoder] | None = None,
        response_timeout: float = 180,
    ) -> AbstractAsyncContextManager[RPC]:
        url = self.options["url"]
        errors_map_ = errors_map or {}

        async def listener(connection: ClientConnection, handlers: AbcTransport):
            try:
                async for message in connection:
                    handlers.on_message(message)
            except asyncio.CancelledError:
                handlers.on_close(None)
            except Exception as exc:
                handlers.on_close(exc)

        @asynccontextmanager
        async def connect_context():
            try:
                listener_task = None
                async with connect(url, additional_headers=additional_headers) as connection:

                    class Transport(AbcTransport):
                        errors_map = dict(**errors_map_)

                        async def send(self, message: str):
                            await connection.send(message)

                    transport = Transport()
                    rpc = RPC(transport, json_encoder, response_timeout)
                    listener_task = asyncio.create_task(listener(connection, transport))
                    yield rpc
            finally:
                if listener_task and not listener_task.done():
                    listener_task.cancel()
                    await listener_task

        return connect_context()



import pytest

from celestia.node_api import Client
from celestia.types.blob import Blob
from celestia.types.rawshare import SampleCoords, RawSample


@pytest.mark.asyncio
async def test_share(node_provider):
    bridge, auth_token = await node_provider("bridge-0")
    client = Client(port=bridge.port["26658/tcp"])
    async with client.connect(auth_token) as api:
        result = await api.blob.submit(
            Blob(b"abc", b"0123456789"),
            Blob(b"abc", b"QWERTYUIOP"),
            Blob(b"xyz", b"ASDFGHJKL"),
        )

        eds = await api.share.get_eds(result.height)
        empty_gnd = await api.share.get_namespace_data(result.height, b"abcd")
        assert empty_gnd == []
        gnd = await api.share.get_namespace_data(result.height, b"abc")
        range_data = await api.share.get_range(result.height, 1, 2)
        samples = await api.share.get_samples(
            (await api.header.get_by_height(result.height)), [SampleCoords(row=0, col=1)]
        )
        coords_data = await api.share.get_share(result.height, 0, 1)
        if samples and isinstance(samples[0], RawSample):
            assert (
                coords_data
                == samples[0].share.data
                == range_data.proof.data[0]
                == eds.data_square[1]
                == gnd[0].shares[0]
            )
        else:
            assert (
                coords_data
                == samples[0]
                == range_data.proof.data[0]
                == eds.data_square[1]
                == gnd[0].shares[0]
            )
        await api.share.get_available(result.height)



import asyncio

import pytest

from celestia.node_api import Client
from celestia.types.blob import Blob
from celestia.types.common import Namespace


@pytest.mark.asyncio
async def test_send_blobs(node_provider):
    bridge, auth_token = await node_provider("bridge-0")
    client = Client(port=bridge.port["26658/tcp"])
    async with client.connect(auth_token) as api:
        blobs = await api.blob.get_all(1, b"abc")
        assert blobs == []

        result = await api.blob.submit(
            Blob(b"abc", b"0123456789"),
            Blob(b"abc", b"QWERTYUIOP"),
            Blob(b"xyz", b"ASDFGHJKL"),
        )

        assert len(result.commitments) == 3

        blobs = await api.blob.get_all(result.height, b"abc")
        assert len(blobs) == 2
        assert blobs[0].data == b"0123456789"
        assert blobs[0].commitment == result.commitments[0]

        assert blobs[1].data == b"QWERTYUIOP"
        assert blobs[1].commitment == result.commitments[1]

        blobs = await api.blob.get_all(result.height, b"xyz")
        assert len(blobs) == 1
        assert blobs[0].data == b"ASDFGHJKL"
        assert blobs[0].commitment == result.commitments[2]

        blob = await api.blob.get(1, b"abc", b"ASDFGHJKL")
        assert blob is None

        blob = await api.blob.get(result.height, b"abc", result.commitments[1])
        assert len(blobs) == 1
        assert blob.data == b"QWERTYUIOP"

        proof = await api.blob.get_proof(1, b"abc", b"ASDFGHJKL")
        assert proof == []

        proof = await api.blob.get_proof(result.height, b"abc", result.commitments[1])
        assert proof

        included = await api.blob.included(result.height, b"xyz", proof, result.commitments[1])
        assert not included

        included = await api.blob.included(result.height, b"abc", proof, result.commitments[1])
        assert included

        com_proof = await api.blob.get_commitment_proof(
            result.height, b"abc", result.commitments[1]
        )
        assert com_proof is not None

        com_proof = await api.blob.get_commitment_proof(
            result.height, b"abc", result.commitments[1]
        )
        assert com_proof is not None


@pytest.mark.asyncio
async def test_blob_exceptions(node_provider):
    bridge, auth_token = await node_provider("bridge-0")
    client = Client(port=bridge.port["26658/tcp"])
    async with client.connect(auth_token) as api:
        with pytest.raises(ValueError):
            await api.blob.get_all(18446744073709551615, b"abc")
        with pytest.raises(ValueError):
            await api.blob.get_all(0, b"abc")
        with pytest.raises(ValueError):
            await api.blob.get_all(-1, b"abc")

        with pytest.raises(ValueError):
            await api.blob.get(18446744073709551615, b"abc", b"ASDFGHJKL")
        with pytest.raises(ValueError):
            await api.blob.get(0, b"abc", b"ASDFGHJKL")
        with pytest.raises(ValueError):
            await api.blob.get(-1, b"abc", b"ASDFGHJKL")

        with pytest.raises(ValueError):
            await api.blob.get_proof(18446744073709551615, b"abc", b"ASDFGHJKL")
        with pytest.raises(ValueError):
            await api.blob.get_proof(0, b"abc", b"ASDFGHJKL")
        with pytest.raises(ValueError):
            await api.blob.get_proof(-1, b"abc", b"ASDFGHJKL")

        with pytest.raises(TypeError):
            Blob(b"abc", None)
        with pytest.raises(ValueError):
            Blob(123, b"ASDFGHJKL")


@pytest.mark.asyncio
async def test_blob_subscribe(node_provider):
    bridge, auth_token = await node_provider("bridge-0")
    client = Client(port=bridge.port["26658/tcp"])

    result = []

    async def submitter(api):
        for i in range(5):
            if i == 2:
                await api.blob.submit(Blob(b"abc", f"QWERTY{i}".encode()))
            else:
                await api.blob.submit(Blob(b"qwe", f"QWERTY{i}".encode()))

    async with client.connect(auth_token) as api:
        submitter_tack = asyncio.create_task(submitter(api))
        try:
            async with asyncio.timeout(30):
                async for blob in api.blob.subscribe(b"qwe"):
                    result.append(blob)
                    if len(result) == 4:
                        break
        finally:
            submitter_tack.cancel()

    assert len(result) == 4
    assert tuple(item.blobs[0].namespace for item in result) == (Namespace(b"qwe"),) * 4



import pytest
from celestia._celestia import types  # noqa

from celestia.types.common import Namespace


@pytest.mark.asyncio
async def test_namespace():
    namespace = Namespace(b"Alesh")
    assert namespace == b"\x00" * 24 + b"Alesh"
    assert str(namespace) == "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAQWxlc2g="
    assert namespace == Namespace("AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAQWxlc2g=")
    with pytest.raises(ValueError):
        Namespace(b"111111111111111111111111111111111")
    with pytest.raises(ValueError):
        Namespace(b"1111111111111111111111111111111111111111111111111111111111")


@pytest.mark.asyncio
async def test_blob():
    blob = types.normalize_blob(b"Alesh", b"0123456789ABCDEF")
    assert blob == {
        "namespace": b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00Alesh",
        "commitment": b"\x88e\rh\xc1\x02\xbd\xfc\xbcc\xa3\xcc\x10\n5\xdf\xcbCh\xa3m\x04\xe1\xeds(\xdf}j>\xab/",
        "data": b"0123456789ABCDEF",
        "index": None,
        "share_version": 0,
    }



import json
from contextlib import asynccontextmanager, AbstractAsyncContextManager
from dataclasses import is_dataclass, asdict
from urllib.parse import urlparse

import _jsonrpc
from _jsonrpc import RPC
from celestia.types import Base64
from .blob import BlobClient
from .das import DasClient
from .fraud import FraudClient
from .header import HeaderClient
from .p2p import P2PClient
from .share import ShareClient
from .state import StateClient


class NodeAPI:
    """Celestia node API"""

    def __init__(self, rpc: RPC):
        self._rpc = rpc

    @property
    def state(self):
        return StateClient(self._rpc)

    @property
    def blob(self):
        return BlobClient(self._rpc)

    @property
    def header(self):
        return HeaderClient(self._rpc)

    @property
    def p2p(self):
        return P2PClient(self._rpc)

    @property
    def das(self):
        return DasClient(self._rpc)

    @property
    def fraud(self):
        return FraudClient(self._rpc)

    @property
    def share(self):
        return ShareClient(self._rpc)


class JSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if is_dataclass(obj):
            return asdict(obj)
        if isinstance(obj, Base64):
            return str(obj)
        return super().default(obj)


class Client(_jsonrpc.Client):
    """Celestia Node API client"""

    def __init__(
        self,
        url: str = None,
        /,
        *,
        auth_token: str = None,
        host: str = "localhost",
        port: int = 26658,
    ):
        pr = urlparse(url or f"ws://{host}:{port}")
        if pr.scheme not in ("ws", "wss"):
            raise ValueError("Unsupported URL scheme, must be ws or wss")
        url = f"{pr.scheme}://{pr.hostname}:{pr.port or port}"
        super().__init__(url, auth_token=auth_token)
        self.errors_map = dict(
            (rpc_err, ValueError)
            for rpc_err in [
                "unmarshaling params",
                "equal to 0",
                "given height is from the future",
                "invalid range",
                "height must be bigger than zero",
                "dial to self attempted",
                "gater disallows connection to peer",
                "notfound desc = delegation with delegator",
                "unknown desc = failed to execute message; message index: 0: invalid shares amount:",
                "cannot redelegate to the same validator",
                "too many unbonding delegation entries for (delegator, validator) tuple",
                "redelegation not found for delegator address",
                "too many redelegation entries for (delegator, src-validator, dst-validator)",
                "datastore: key not found",
                "reserved namespace",
                "invalid namespace length",
                "invalid data size",
                "blob size mismatch",
                "unsupported share version",
                "zero blob size",
                "no blobs",
                "invalid blob signer",
                "invalid namespace type",
            ]
        )

    def connect(
        self, auth_token: str = None, /, response_timeout: float = 180, **_
    ) -> AbstractAsyncContextManager[NodeAPI]:
        """Creates and return connection context manager."""
        headers = []
        if auth_token := auth_token or self.options.get("auth_token"):
            headers.append(("Authorization", f"Bearer {auth_token}"))

        @asynccontextmanager
        async def connect_context():
            async with super(Client, self).connect(
                additional_headers=headers,
                errors_map=self.errors_map,
                response_timeout=response_timeout,
                json_encoder=JSONEncoder,
            ) as rpc:
                yield NodeAPI(rpc)

        return connect_context()



import asyncio
from dataclasses import asdict

import pytest
from celestia._celestia import types as ext  # noqa

from celestia.node_api import Client
from celestia.types import Namespace
from celestia.types.blob import Blob


def test_create_blob():
    blob = Blob(b"abc", b"0123456789")
    assert dict((k, str(v)) for k, v in asdict(blob).items()) == {
        "commitment": "8M4WkNbOtjABYe0ymm4Q/BgfrBmDT4FpCXRrviKFrIE=",
        "data": "MDEyMzQ1Njc4OQ==",
        "index": "None",
        "namespace": "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAABhYmM=",
        "share_version": "0",
        "signer": "None",
    }

    assert Blob(**asdict(blob))

    signed_blob = Blob(
        b"abc", b"0123456789", signer="celestia1qypqxpq9qcrsszg2pvxq6rs0zqg3yyc5wgawu3"
    )
    assert dict((k, str(v)) for k, v in asdict(signed_blob).items()) == {
        "commitment": "pk39MJeTRHg0FuAZX0PKAnxQLJ/PAprwBrzgGNJ2hu0=",
        "data": "MDEyMzQ1Njc4OQ==",
        "index": "None",
        "namespace": "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAABhYmM=",
        "share_version": "1",
        "signer": "AQIDBAUGBwgJCgsMDQ4PEBESExQ=",
    }

    assert Blob(**asdict(signed_blob))

    assert signed_blob.commitment != blob.commitment


def test_blob_errors():
    blob = Blob(b"abc", b"0123456789")
    signed_blob = Blob(
        b"abc", b"0123456789", signer="celestia1qypqxpq9qcrsszg2pvxq6rs0zqg3yyc5wgawu3"
    )

    with pytest.raises(ValueError, match="Wrong namespaces"):
        Blob(b"abc0123456789", b"0123456789")

    with pytest.raises(ValueError, match="Invalid namespace size"):
        Blob(b"abc0123456789", b"0123456789")

    with pytest.raises(ValueError, match="Wrong commitment"):
        Blob(**dict(asdict(blob), commitment=b"0123456789"))

    with pytest.raises(ValueError, match="Invalid address"):
        Blob(b"abc", b"0123456789", signer="celestiaXXX=")

    with pytest.raises(ValueError, match="Wrong commitment"):
        Blob(**dict(asdict(signed_blob), commitment=b"0123456789"))

    with pytest.raises(ValueError, match="Wrong share version"):
        Blob(**dict(asdict(signed_blob), share_version=0))


async def test_sending_signed_blob(node_provider):
    bridge, auth_token = await node_provider("bridge-0")
    client = Client(port=bridge.port["26658/tcp"])
    async with client.connect(auth_token) as api:
        # unsigned
        blob = Blob(b"abc", b"0123456789")
        result = await api.blob.submit(blob)
        assert len(result.commitments) == 1
        assert result.commitments[0] == blob.commitment

        blobs = await api.blob.get_all(result.height, b"abc")
        assert len(blobs) == 1
        assert blobs[0].commitment == blob.commitment
        assert blobs[0].data == b"0123456789"
        assert blobs[0].share_version == 0

        # signed
        signer_address = await api.state.account_address()
        signed_blob = Blob(b"abc", b"0123456789", signer=signer_address)

        result = await api.blob.submit(signed_blob)
        assert len(result.commitments) == 1
        assert result.commitments[0] == signed_blob.commitment

        blobs = await api.blob.get_all(result.height, b"abc")
        assert blobs[0].commitment == signed_blob.commitment
        assert ext.bytes2address(blobs[0].signer) == signer_address
        assert blobs[0].data == b"0123456789"
        assert blobs[0].share_version == 1
        assert len(blobs) == 1

        # bad signer
        bad_signer_blob = Blob(
            b"abc", b"0123456789", signer="celestia1qypqxpq9qcrsszg2pvxq6rs0zqg3yyc5wgawu3"
        )
        with pytest.raises(ValueError, match="invalid blob signer"):
            await api.blob.submit(bad_signer_blob)

        # auto signer
        blob = Blob(b"abc", b"0123456789")
        assert blob.share_version == 0
        result = await api.blob.submit(blob, signer_address=True)
        assert len(result.commitments) == 1
        blobs = await api.blob.get_all(result.height, b"abc")
        assert blobs[0].commitment != blob.commitment
        assert ext.bytes2address(blobs[0].signer) == signer_address
        assert blobs[0].data == b"0123456789"
        assert blobs[0].share_version == 1
        assert len(blobs) == 1

        # bad auto signer
        bad_signer_blob = Blob(b"abc", b"0123456789")
        with pytest.raises(
            ValueError, match="`signer_address` does not match the current account address"
        ):
            await api.blob.submit(
                bad_signer_blob, signer_address="celestia1qypqxpq9qcrsszg2pvxq6rs0zqg3yyc5wgawu3"
            )

        # bad BLOB signer auto
        bad_signer_blob = Blob(
            b"abc", b"0123456789", signer="celestia1qypqxpq9qcrsszg2pvxq6rs0zqg3yyc5wgawu3"
        )
        with pytest.raises(
            ValueError, match="`signer_address` of blob does not match the current account address"
        ):
            await api.blob.submit(bad_signer_blob, signer_address=True)


async def test_blob_subscribe(node_provider):
    bridge, auth_token = await node_provider("bridge-0")
    client = Client(port=bridge.port["26658/tcp"])

    result = []

    async def submitter(api):
        for i in range(5):
            if i == 2:
                await api.blob.submit(Blob(b"abc", f"QWERTY{i}".encode()), signer_address=True)
            else:
                await api.blob.submit(Blob(b"qwe", f"QWERTY{i}".encode()), signer_address=True)

    async with client.connect(auth_token) as api:
        account_address = await api.state.account_address()
        submitter_tack = asyncio.create_task(submitter(api))
        try:
            async with asyncio.timeout(30):
                async for blob in api.blob.subscribe(b"qwe"):
                    result.append(blob)
                    if len(result) == 4:
                        break
        finally:
            submitter_tack.cancel()

    assert len(result) == 4
    assert tuple(item.blobs[0].namespace for item in result) == (Namespace(b"qwe"),) * 4
    assert (
        tuple(ext.bytes2address(item.blobs[0].signer) for item in result) == (account_address,) * 4
    )
    assert all(item.blobs[0].share_version == 1 for item in result)
